from setuptools import setup, find_packages
setup(
name='rtsvg',
version='0.1.15',
author='D. Trimm',
author_email='dave.trimm@gmail.com',
description='Visualization Package Leveraging SVG in Jupyter Notebooks',
packages=['rtsvg'],
classifiers=[
'Programming Language :: Python :: 3',
'License :: OSI Approved :: Apache Software License',
'Operating System :: OS Independent',
],
python_requires='>=3.11',
url='https://github.com/datrcode/racetrack_svg_framework'
)

